from . import decorator
from . import filehandlers
from . import configuration 
from . import articles
from . import preprocessor
from bottle import default_app, route, run, template, static_file, request, auth_basic
import jinja2
import os
import markdown
import logging
from string import Template
import difflib


def make_sure_dir_exists(folder_path):
    dir = os.path.dirname(folder_path)
    if not os.path.exists(dir):
        logging.info("Creating directory: %s", dir)
        try:
            os.makedirs(dir)
        except:
            logging.warning("Cannot create directory: %s", dir)

def make_embedded_articles(template):
    r = {}
    if template.embedd != None:
        for variable_name, article in template.embedd.items():
            logging.info("Embedding article: %s", article)
            r[variable_name] = markdown.markdown(articles.get_article_from(configuration['source'], article).content)
    return r

def get_ratio(string1, string2):
    ratio = difflib.SequenceMatcher(None, string1, string2).ratio()
    return ratio

def get_matching_content(string1, string2):
    occur = string2.count(string1)
    if occur == 1:
        return string1

    src_len = len(string1)

    last_ratio = None

    for i in range(0, len(string2)):
        n = string2[i:(i+src_len)]
        r = get_ratio(string1, n)
        if r > 0.9:
            if last_ratio is not None:
                if r > last_ratio[0]:
                    last_ratio = (r, n, i)
                else:
                    return string2[last_ratio[2]:last_ratio[2]+src_len+20]
            else:
                last_ratio = (r, n, i)
        else:
            if last_ratio is not None:
                return string2[last_ratio[2]:last_ratio[2]+src_len+20]
    return None

def replace_string(file_path, before, after):
    with open(file_path, 'r') as f:
        file_contents = f.read()

    occur = file_contents.count(before)

    if (occur != 1):
        logging.error("More than one match found : %d", occur);
        return False
    else:
        with open(file_path, 'w') as f:
            new_contents = file_contents.replace(before, after)
            f.write(new_contents)
            return True

def generate_article(decorator, environment, article):
    template = decorator.get_template_for_type(article.type)
    if template == None:
        raise Exception("No template found for article type: %s", article.type)
    jinja_template = environment.get_template(template.file)
    md_content = preprocessor.process(article.content, article.name)
    content = markdown.markdown(md_content, extensions=['markdown.extensions.toc', 'markdown.extensions.fenced_code'])
    data = article.data
    data.update(make_embedded_articles(template))
    return jinja_template.render(content=content, **data).encode("utf-8")

def generate(source, output, decorator_path, templates_folder, template_map):
    d = decorator.Decorator(decorator_path, templates_folder, template_map)
    d.add_handler('.less', filehandlers.LessProcessor)
    d.build_decoration(output)
    environment = jinja2.Environment(loader=jinja2.FileSystemLoader(d.template_folder))
    permalinks = {}
    for article in articles.get_articles_from(source):
        if article.permalink != None:
            if article.permalink in permalinks:
                raise Exception("Duplicate permalink found %s's permalink and %s's permalinks are same" % (article.name, permalinks[article.permalink]))
            permalinks[article.permalink] = article.name
            output_path = os.path.join(output, *os.path.split(article.permalink))
            if article.publish:
                make_sure_dir_exists(output_path)
                with open(output_path, 'wb') as f:
                    try:
                        f.write(generate_article(d, environment, article))
                    except:
                        logging.warning("Error generating article: %s", article.name)
                        pass
                    logging.info("File written: %s", output_path)
            else:
                logging.info("Not publishing %s",  article.name)
                try:
                    os.remove(output_path)
                except:
                    pass

def inject_scripts(article):
    return Template('''
    <script src="//code.jquery.com/jquery-3.3.1.min.js"></script>
    <script>
        var permalink = '${permalink}';
        var before = "";
        function openDialog(e, file_contents) {
            if ($$('#save-button').length > 0) {
                return;
            }
            var je = $$(e);
            before = je.html();
            var saveB = $$('<button>').insertAfter(je).html("Save");
            var cancelB = $$('<button>').insertAfter(saveB).html("Cancel");
            saveB.attr('id', "save-button");
            cancelB.attr('id', "cancel-button");
            cancelB.click(function() {
                je.html(before);
                je.attr('contenteditable', "false");
                saveB.remove();
                cancelB.remove();
            });

            saveB.click(function() {
                $$.post('/feni_rt_save_/'+ permalink, {before: file_contents, after: je.text()}, function(response) {
                    if (response == 'failed') {
                        je.html(before);
                    }
                    je.attr('contenteditable', "false");
                    saveB.remove();
                    cancelB.remove();
                    location.reload();
                });
            });
            je.text(file_contents);
            je.attr('contenteditable', 'true');
            je.focus();
        }
        $$(function() {
            $$('p, h1, h2, h3, h4, h5, h6, blockquote, b, i, strong, em, small, li').click(function(e) {
                var je = $$(e.target);
                $$.post('/feni_rt_get_/'+permalink, {before: je.text()}, function(file_content) {
                    openDialog(e.target, file_content);
                });
            });
        });

    </script>''').substitute({'permalink': article.permalink})

template_html =  '''
<!DOCTYPE html>
<html lang="en">

<head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <meta name="description" content="">
    <meta name="author" content="">
    <link href="/styles.css" rel="stylesheet" type="text/css"></link>
</head>

<body class="language-markup">
    <% for item in items: %>
    <div><a href="feni_rt_edit_/{{item.permalink}}">{{ item.name }}</a></div>
    <% end %>
</body>

</html>'''

@route('/feni_rt_get_/<permalink:path>', method='POST')
def get_matching(permalink):
    before = request.forms.get('before')
    config = configuration.read()
    for article in articles.get_articles_from(config['source']):
        if article.permalink == permalink:
            with open(article.path, 'r') as f:
                file_contents = f.read()
                m = get_matching_content(before, file_contents)
                if m is not None:
                    return m
                else:
                    return template("{{content}}", content="")

@route('/feni_rt_save_/<permalink:path>', method='POST')
def save(permalink):
    before = request.forms.get('before')
    after = request.forms.get('after')
    config = configuration.read()
    for article in articles.get_articles_from(config['source']):
        if article.permalink == permalink:
            if replace_string(article.path, before, after):
                return template("{{content}}", content="success")
    return template("{{content}}", content="failed")

@route('/feni_rt_edit_/<permalink:path>')
def edit(permalink):
    config = configuration.read()
    d = decorator.Decorator(config['decorator'], config['template'],  config['templatemap'])
    d.add_handler('.less', filehandlers.LessProcessor)
    environment = jinja2.Environment(loader=jinja2.FileSystemLoader(d.template_folder))
    for article in articles.get_articles_from(config['source']):
        if article.permalink == permalink:
            article_html = generate_article(d,environment,article)
            article_html = article_html.replace(str.encode("</head>"), str.encode(inject_scripts(article) + "</head>"))
            return article_html
    return template("Article not found!")

@route('/')
def index():
    config = configuration.read()
    items = []
    for article in articles.get_articles_from(config['source']):
        items.append(article)
    return template(template_html, items=items)

@route('/<filename:path>', name='static')
def serve_static(filename):
    config = configuration.read()
    return static_file(filename, root=config['decorator'])

def main():
    try:
        config = configuration.read()
        if config['server']:
            run(host='0.0.0.0')
        generate(config['source'], config['output'], config['decorator'], config['template'], config['templatemap'])
        logging.info("Successfully created site in %s", config['output'])
    except Exception as ex:
        logging.error(str(ex))
        raise

app = default_app()



import os
import glob
import yaml
import logging
from .  import exceptions

class Article:
    def __init__(self, name, path, header, content):
        self.name = name
        self.path = path
        self.type = header['type']
        self.permalink = header['permalink'] if 'permalink' in header else None
        self.main = header['main'] if 'main' in header else True
        self.data = header['data']
        self.publish = header['publish'] if 'publish' in header else False
        self.content = content

def get_articles_from(folder, encoding = 'utf-8'):
    paths = os.path.join(folder, '*.md')
    for filename in glob.iglob(paths):
        logging.info("Found article: %s",filename)
        yield get_article_from(folder, filename, encoding=encoding)
                
def get_article_from(folder, filename, encoding = 'utf-8'):
    path = os.path.join(folder, filename)
    with open(path, 'r', encoding=encoding) as f:
        segments = split_file(f)
        if (segments[0] != None):
            header = yaml.load(segments[0])
            header['data'] = get_data(header)
            return Article(os.path.basename(path), path, header, segments[1])
        else:
            raise exceptions.EmptyArticleError(filename)

def get_data(header):
    r = {}
    for key in header:
        if key[0:5] == 'data-':
            r[key[5:]] = header[key]
    return r

def split_file(f):
    segments = []
    current = ''
    for line in f:
        if line.strip() == '---':
            segments.append(current)
            current = ''
        else:
            current += line
    if current != '':
        segments.append(current)
    elif len(segments) == 0:
        raise exceptions.EmptyFileError
    if len(segments) == 1:
        segments.insert(0, None)
    return segments

            

        



from . import exceptions
import os
import yaml
import argparse
import logging

def find_configuration_file_from(folder, filename):
    full_folder_path = os.path.abspath(folder)
    full_file_name = os.path.join(full_folder_path, filename)
    if os.path.isfile(full_file_name):
        return full_file_name
    parent_folder = os.path.dirname(full_folder_path)
    if parent_folder != folder:
        return find_configuration_file_from(parent_folder, filename)
    raise exceptions.FileNotFoundError(filename)

def read_configuration(file_path):
    file_path = os.path.realpath(file_path)
    with open(file_path, 'r') as f:
        raw_conf = yaml.load(f)
        parent_folder = os.path.dirname(file_path)
        conf = {
                'source': convert_path(parent_folder, raw_conf['source'], True),
                'decorator': convert_path(parent_folder, raw_conf['decorator'], True),
                'output': convert_path(parent_folder, raw_conf['output']),
                'server': False
        }
        if 'template' in raw_conf:
            conf['template'] = convert_path(parent_folder, raw_conf['template'], True)
        else:
            conf['template'] = convert_path(os.path.dirname(conf['decorator']), 'templates', True)
        if 'templatemap' in raw_conf:
            conf['templatemap'] = os.path.join(parent_folder, raw_conf['templatemap'])
        else:
            conf['templatemap'] = os.path.join(conf['template'], 'templates.yaml')
        return conf

def convert_path(folder, filename, check = False):
    full_path = os.path.join(folder, filename)
    if check:
        if os.path.isdir(full_path):
            return full_path
        else:
            raise FileNotFoundError(full_path)
    else:
        return full_path

def read():
    parser = argparse.ArgumentParser(description='Feni static site generator!', formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--source', help='Source directory for articles')
    parser.add_argument('--server', action='store_true', help='Serve pages via local server')
    parser.add_argument('--decorator', help='Decorator directory')
    parser.add_argument('--output', help='Output directory')
    parser.add_argument('--template', help='Template directory')
    parser.add_argument('--templatemap', help='Template map file')
    parser.add_argument('--config', help='Use specified configuration file \n %s' % config_help)
    parser.add_argument('--verbose', action='store_true', help='Verbose output')
    arguments = parser.parse_args()
    if arguments.config:
        configuration = read_configuration(arguments.config)
    else:
        try:
            config_file = find_configuration_file_from('.', 'feni.yaml')
        except exceptions.FileNotFoundError:
            configuration = {}
        else:
            configuration = read_configuration(config_file)
    if arguments.source:
        configuration['source'] = os.path.abspath(arguments.source)
    if arguments.decorator:
        configuration['decorator'] = os.path.abspath(arguments.decorator)
    if arguments.output:
        configuration['output'] = os.path.abspath(arguments.output)
    if arguments.template:
        configuration['template'] = os.path.abspath(arguments.template)
    else:
        if 'template' not in configuration:
            if 'decorator' in configuration:
                configuration['template'] = convert_path(configuration['decorator'], 'templates', True)
            else:
                raise Exception("Cannot figure out the directory to read templates from")
    if arguments.templatemap:
        configuration['templatemap'] = os.path.abspath(arguments.templatemap)
    else:
        if 'templatemap' not in configuration:
            configuration['templatemap'] = os.path.join(configuration['template'], 'templates.yaml')
    if arguments.verbose:
        logging.basicConfig(level=logging.DEBUG)
    if arguments.server:
        configuration['server'] = True
    for k in ['source', 'decorator', 'output']:
        if k not in configuration:
            raise exceptions.RequiredValueMissing(k)
    if not os.path.isdir(configuration['source']):
        raise exceptions.FileNotFoundError(configuration['source'])
    if not os.path.isdir(configuration['decorator']):
        raise exceptions.FileNotFoundError(configuration['decorator'])
    if not os.path.isdir(configuration['template']):
        raise exceptions.FileNotFoundError(configuration['template'])
    if not os.path.isfile(configuration['templatemap']):
        raise exceptions.FileNotFoundError(configuration['templatemap'])
    return configuration

config_help = '''
Feni needs the following information to function, that it can read
from command line arguments or from a configuration file.

1. A source directory. This is where Feni looks for content. Right now it only
look for .md files. Each of this file should be of format as shown below.

    main: true
    type: post
    permalink: test-post.html
    data-title: My test post
    data-subtitle: This is as test post
    ---
    ###This is the content of the post in markdown format###

Every article needs to have a header in yaml format that contain information
regarding how the final html should be generated. The first one, the 'main'
attribute denotes if the article is intended to be a whole article or if it is
part of a bigger article. If it is a whole article, this attribute should be true.
If it is a sub article, then this should be false.

The next attribute, the 'type', as the name implies, denotes the type of the article.
This attribute is used to find a template that can be used to expand this article.

'permalink' attribute is the final location of the generated html file. This should be
relative path. The generated html is placed under this path relative to the output
directory. 

'publish' attribute is used to denote wether to generate an article or not. If the
'publish' status of an article is changed to false, then that article is removed
from it's permalink location in the target folder, if it already exists.

the data- attributes are custom fields that you can define, which will be available in
your template. This can be any valid yaml value. For example, you this can be a list
where each item is information about an article. This list can be used to display a
list of articles in the target template. If you define a value 'data-article_name', then
that value will be available in the template as 'article_name' key.

2. A destination directory. This is where the generated files will be put.

3. A decorator directory. This is where all the other files, like css files, js files,
background images come from. This folder will be mostly copied as such into the output
folder. If there are files matching the enabled file handlers, then those will be passed
through them. Right now only .less files will be processed into css files.

4. A path to a templates folder that should contain all the jinja templates.

5. Path to a template configuration file. This file has the following format.

about.html:
    types:
        - aboutme
post.html:
    types:
        - post

As you can see, this just lists the template files and declares the type of
articles it can handle. By default, Feni looks for this file in the templates folder
under the name templates.yaml. But you can change that using the --templatemap
option

Feni can read this from a configuration file in yaml format.

source: articles
destination: output
decorator: design

By default, Feni looks for the this file in the current folder and folders above it.
Paths read from it will be considers as relative to the directory where it is found.
Template paths in template.yaml will be taken relative to the template directory.

'''



class FileNotFoundError(Exception):
    def __init__(self, message):
        super().__init__("File is not found:"+message)

class EmptyFileError(Exception):
    def __init__(self, message):
        super().__init__("File is empty:"+message)

class EmptyArticleError(Exception):
    def __init__(self, message):
        super().__init__("Article is empty:"+message)

class RequiredValueMissing(Exception):
    def __init__(self, message):
        super().__init__("Required value "+message+" was missing")



import os
import shutil
import yaml
import pathlib
import logging

def check_if_child(parent, child):
    if parent == child:
        return True
    for child_parent in child.parents:
        if child_parent == parent:
            return True
    return False

class Template:
    def __init__(self, template_file, embedd):
        self.file = template_file
        self.embedd = embedd

class Decorator:
    def __init__(self, folder_path, templates_folder, template_map):
        self.folder = folder_path
        self.template_folder = templates_folder
        self.file_handlers = {}
        with open(template_map, 'r') as f:
            self.template_configuration = yaml.load(f)
            
    def add_handler(self, extension, handler):
        self.file_handlers[extension] = handler

    def get_template_for_type(self, type):
        for template_file, configuration in self.template_configuration.items():
            if type in configuration['types']:
                return Template(template_file, configuration['embedd'] if 'embedd' in configuration else None)

    def build_decoration(self, target):
        logging.info("Buliding decoration")
        source_path = pathlib.PurePath(self.folder)
        templates_path = pathlib.PurePath(self.template_folder)
        if not os.path.isdir(target):
            os.mkdir(target)
        for dirinfo in os.walk(self.folder):
            for dir in dirinfo[1]:
                p = pathlib.PurePath(dirinfo[0], dir)
                if check_if_child(templates_path, p):
                    continue
                destination = os.path.join(target, str(p.relative_to(source_path)))
                if not os.path.isdir(destination):
                    os.mkdir(destination)
            for file in dirinfo[2]:
                p = pathlib.PurePath(dirinfo[0], file)
                if check_if_child(templates_path, p):
                    continue
                this_source = str(p)
                destination = os.path.join(target, str(p.relative_to(source_path)))
                if p.suffix in self.file_handlers:
                    handler = self.file_handlers[p.suffix]
                    handler.process_file(str(p), destination)
                else:
                    shutil.copyfile(this_source, destination)



from setuptools import setup, find_packages

setup(
    name = "Feni",
    description='A simple static site generator.',
    url='http://bitbucket.org/sras/feni',
    author='Sandeep.C.R',
    author_email='sandeepcr2@gmail.com',
    license='MIT',
    version = "1.12",
    packages = ['feni'],
    entry_points = {
        "console_scripts":['feni=feni.feni:main']
    },
    install_requires=[
      'lesscpy',
      'markdown',
      'pyyaml',
      'bottle'
    ]
)



import lesscpy
import pathlib
import os
import logging

class LessProcessor:
    def process_file(filename, destination):
        old_path = os.getcwd()
        try:
            os.chdir(os.path.dirname(filename))
            destination = str(pathlib.PurePath(destination).with_suffix('.css'))
            with open(destination, 'w') as d:
                d.write(lesscpy.compile(filename))
        except Exception as e:
            logging.warning("Exception %s" % str(e))
            os.chdir(old_path)
